/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;

import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.plugin.AbstractUIPlugin;

import com.ibm.etools.c.plugin.CPlugin;

/**
 * CPreferenceStore is a place to store the user preference options for the C
 * Importer. The preferences come from two sources: the UI Preferences Page and
 * the options argument passed to the CImporter.load() method.
 * 
 * The options in the UI Preferences Page include compiler and storage layout
 * information. They are mostly used to control the population of the Type
 * Descriptor model. There are also some advanced options which can only be set
 * in the options argument passed to the CImporter.load() method. They affect
 * which objects are included in the generated Extent list. These options are
 * not exposed in the UI Preferences Page because normal users do not need to
 * touch them.
 */
public class CPreferenceStore {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private static CPreferenceStore _instance = new CPreferenceStore();

    // Alignment rule constants
    public static final int NATURAL_ALIGNMENT = 0;

    public static final int POWER_ALIGNMENT = 1;

    public static final int PACKED_ALIGNMENT = 2;

    private static final String LITTLE_ENDIAN = "Little Endian";

    private static final String BIG_ENDIAN = "Big Endian";

    private static final String NATURAL = "Natural";

    private static final String POWER = "Power";

    private static final String PACKED = "Packed";

    private static final String SPACE = "SPACE";

    private static final String NULL = "NULL";

    private static final String TRUE = "TRUE";

    private static final String FALSE = "FALSE";

    private static final String empty = "";

    /*
     * These options can be set in the UI Prefernces Page. They can also be set
     * in key-value pairs passed to the options argument.
     */
    private static String _compilerName; // Compiler name

    private static String _compilerFlags; // Compiler flags

    private static String _includePaths; // Include paths

    // Is the target platform big endian or little endian? (default: little
    // endian)
    private static boolean _isBigEndian = false;

    // Is the address size 64 bit or 32 bit? (default: 32 bit)
    private static boolean _is64bitAddress = false;

    // Is the size of long double 128 bit or 64 bit? (default: 64 bit)
    private static boolean _isLongDouble128 = false;

    //Are unsigned short arrays DBSC strings?(default: NO)
    private static boolean _isTreatUShortArrayAsDBCS = false;

    // The alignment rule (default: natural alignment)
    private static int _alignment = NATURAL_ALIGNMENT;

    // The pack level (default: 1)
    private static int _packLevel = 1;

    // The size of enums (default: 4)
    private static int _sizeEnums = 4;

    // The size of wchar (default: 2)
    private static int _sizeWchar = 2;

    // The code page
    private static String _codePage = null;

    // The floating point format
    private static String _floatingPointFormat = null;

    // String representation, null-terminated or space padded (default:
    // null-terminated)
    private static boolean _isStringNullTerminated = true;

    /*
     * These options are used by the Type Descriptor model. They are not
     * provided in the UI, but can still be set from the options argument.
     */
    private static String _operatingSystem = null;

    private static String _hardwarePlatform = null;

    private static String _compilerVersion = null;

    private static String _osVersion = null;

    private static String _paddingCharacter = null;

    /*
     * Advanced settings to control how the Extent is generated. These settings
     * do not exist in the Preference Page.
     */
    // Include global variables in the Extent? (default: no)
    private static boolean _isIncludeGlobalVariables = false;

    // Create an implicit field for a type object? (default: no)
    private static boolean _isCreateImplicitFieldsForTypes = false;

    // Include objects from include files? (default: yes)
    private static boolean _isProcessIncludeFiles = true;

    // Expand the object list? (default: yes)
    private static boolean _isGenerateFullObjectList = true;

    // Include Type Descriptor information? (default: yes)
    private static boolean _isPopulateTD = true;

    // Reuse the parse information from the last importing session? (default:
    // no)
    private static boolean _isReuseParseInfo = false;

    // If this flag is set, the C Importer throws an exception when it
    // encounters a parser error. (default: no)
    private static boolean _isThrowExceptionAtParseError = false;

    // If this flag is set, the C Importer will treat an unresolved type as
    // a parser error. (default: no)
    private static boolean _isTreatUnknownTypeAsError = false;

    /**
     * Create a CPreferenceStore object
     */
    private CPreferenceStore() {
    }

    /**
     * Return the single instance
     */
    public static CPreferenceStore instance() {
        return _instance;
    }

    /**
     * Retrieve the preference settings from the plugin preference store. This
     * method is called once per importer session.
     */
    public static void getValuesFromPlugin() {

        resetValues();
        if (Platform.isRunning()) {
            AbstractUIPlugin plugin = CPlugin.getDefault();
            if (plugin != null && PlatformUI.isWorkbenchRunning()) {
                IPreferenceStore store = plugin.getPreferenceStore();
                if (store != null) {
                    _compilerName = new String(store.getString("com.ibm.etools.c.C_COMPILER_NAME"));
                    _compilerFlags = new String(store.getString("com.ibm.etools.c.C_COMPILER_FLAGS"));
                    _codePage = new String(store.getString("com.ibm.etools.c.C_CODEPAGE"));
                    _floatingPointFormat = new String(store.getString("com.ibm.etools.c.C_FLOATING_POINT_FORMAT"));

                    String stringEncoding = store.getString("com.ibm.etools.c.C_STRING_ENCODING");

                    if (stringEncoding != null && (stringEncoding.equals(NULL) || stringEncoding.length() <= 0)) {
                        _isStringNullTerminated = true;

                    } else {
                        _isStringNullTerminated = false;

                        if (stringEncoding != null && stringEncoding.toUpperCase().equals(SPACE)) {
                            _paddingCharacter = " ";

                        } else {
                            _paddingCharacter = stringEncoding;
                        }
                    }

                    _includePaths = new String(store.getString("com.ibm.etools.c.C_INCLUDE_PATH"));

                    String byteOrder = store.getString("com.ibm.etools.c.C_BYTE_ORDER");
                    if (byteOrder != null && byteOrder.equals(BIG_ENDIAN))
                        _isBigEndian = true;
                    else
                        _isBigEndian = false;

                    int addressSize = store.getInt("com.ibm.etools.c.C_ADDRESS_SIZE");
                    _is64bitAddress = (addressSize == 64);

                    String USArray = store.getString("com.ibm.etools.c.C_US_SHORT_ARRAYS_DBCS");
                    if (USArray != null && USArray.equals(TRUE))
                        _isTreatUShortArrayAsDBCS = true;
                    else
                        _isTreatUShortArrayAsDBCS = false;

                    int sizeofLongDouble = store.getInt("com.ibm.etools.c.C_SIZE_OF_LONG_DOUBLE");
                    _isLongDouble128 = (sizeofLongDouble == 128);

                    String alignmentRule = store.getString("com.ibm.etools.c.C_ALIGNMENT_RULE");
                    if (alignmentRule != null && alignmentRule.equals(POWER))
                        _alignment = POWER_ALIGNMENT;
                    else if (alignmentRule != null && alignmentRule.equals(PACKED))
                        _alignment = PACKED_ALIGNMENT;
                    else
                        _alignment = NATURAL_ALIGNMENT;

                    _packLevel = store.getInt("com.ibm.etools.c.C_PACK_LEVEL");
                    if (_packLevel < 1)
                        _packLevel = 1;

                    // Set the pack level to be a big number if the alignment
                    // rule is not packed
                    // alignment.
                    if (_alignment != PACKED_ALIGNMENT)
                        _packLevel = 128;

                    _sizeEnums = store.getInt("com.ibm.etools.c.C_SIZE_OF_ENUM");
                    if (_sizeEnums < 1)
                        _sizeEnums = 4;

                    _sizeWchar = store.getInt("com.ibm.etools.c.C_SIZE_OF_WCHAR");
                    if (_sizeWchar != 4 && _sizeWchar != 2)
                        _sizeWchar = 2;

                    _isThrowExceptionAtParseError = store.getBoolean("com.ibm.etools.c.C_THROW_EXCEPTION_AT_PARSE_ERROR");
                    _isTreatUnknownTypeAsError = store.getBoolean("com.ibm.etools.c.C_TREAT_UNKNOWN_TYPE_AS_ERROR");
                }
            }
        }

    }

    /**
     * Parse String options
     */
    public static void parseOptionsAsString(String options) {
        if (options == null || options.length() == 0)
            return;

        StringTokenizer tokenizer = new StringTokenizer(options);
        while (tokenizer.hasMoreTokens()) {
            String token = tokenizer.nextToken();
            if (token.toLowerCase().equals("-notd")) {
                _isPopulateTD = false;
            } else if (token.toLowerCase().equals("-noexpand")) {
                _isGenerateFullObjectList = false;
            } else if (token.toLowerCase().equals("-reuseparseinfo")) {
                _isReuseParseInfo = true;
            } else if (token.toLowerCase().equals("-includeglobalvars")) {
                _isIncludeGlobalVariables = true;
            } else if (token.toLowerCase().equals("-createfieldfortype")) {
                _isCreateImplicitFieldsForTypes = true;
            } else if (token.toLowerCase().equals("-noincludes")) {
                _isProcessIncludeFiles = false;
            } else if (token.toLowerCase().equals("-throwexceptionatparseerror")) {
                _isThrowExceptionAtParseError = true;
            } else if (token.toLowerCase().equals("-treatunknowntypeaserror")) {
                _isTreatUnknownTypeAsError = false;
            } else {
                CResource.writeLog(CPreferenceStore.instance(), "parseOptionsAsString", CResource.getString("_ERROR_Invalid_Option", token));
            }
        }
    }

    /**
     * Parse options stored in a HashMap as key-value pairs.
     */
    public static void parseOptionsAsKey(HashMap map) throws Exception {
        Set keySet = map.keySet();
        Iterator it = keySet.iterator();
        while (it.hasNext()) {
            String key = (String) it.next();
            if (key.equals("com.ibm.etools.c.C_COMPILER_NAME"))
                _compilerName = (String) map.get(key);
            else if (key.equals("com.ibm.etools.c.C_COMPILER_FLAGS"))
                _compilerFlags = (String) map.get(key);
            else if (key.equals("com.ibm.etools.c.C_INCLUDE_PATH"))
                _includePaths = (String) map.get(key);
            else if (key.equals("com.ibm.etools.c.C_BYTE_ORDER")) {
                String byteOrder = (String) map.get(key);
                if (byteOrder != null && byteOrder.equals(BIG_ENDIAN))
                    _isBigEndian = true;
                else if (byteOrder != null && byteOrder.equals(LITTLE_ENDIAN))
                    _isBigEndian = false;
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", byteOrder, key));
            } else if (key.equals("com.ibm.etools.c.C_ADDRESS_SIZE")) {
                String addressSize = (String) map.get(key);
                if (addressSize != null && addressSize.equals("64"))
                    _is64bitAddress = true;
                else if (addressSize != null && addressSize.equals("32"))
                    _is64bitAddress = false;
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", addressSize, key));
            }

            else if (key.equals("com.ibm.etools.c.C_US_SHORT_ARRAYS_DBCS")) {
                String ushortArray = (String) map.get(key);
                if (ushortArray != null && ushortArray.toUpperCase().equals(TRUE))
                    _isTreatUShortArrayAsDBCS = true;
                else if (ushortArray != null && ushortArray.toUpperCase().equals(FALSE))
                    _isTreatUShortArrayAsDBCS = false;
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", ushortArray, key));
            }

            else if (key.equals("com.ibm.etools.c.C_SIZE_OF_LONG_DOUBLE")) {
                String sizeOfLongDouble = (String) map.get(key);
                if (sizeOfLongDouble != null && sizeOfLongDouble.equals("128"))
                    _isLongDouble128 = true;
                else if (sizeOfLongDouble != null && sizeOfLongDouble.equals("64"))
                    _isLongDouble128 = false;
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", sizeOfLongDouble, key));
            } else if (key.equals("com.ibm.etools.c.C_ALIGNMENT_RULE")) {
                String alignmentRule = (String) map.get(key);
                if (alignmentRule != null && (alignmentRule.toUpperCase().equals("NATURAL") || alignmentRule.equals("0")))
                    _alignment = NATURAL_ALIGNMENT;
                else if (alignmentRule != null && (alignmentRule.toUpperCase().equals("POWER") || alignmentRule.equals("1")))
                    _alignment = POWER_ALIGNMENT;
                else if (alignmentRule != null && (alignmentRule.toUpperCase().equals("PACKED") || alignmentRule.equals("2")))
                    _alignment = PACKED_ALIGNMENT;
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", alignmentRule, key));
            } else if (key.equals("com.ibm.etools.c.C_PACK_LEVEL")) {
                String packLevel = (String) map.get(key);
                if (packLevel.equals("1") || packLevel.equals("2") || packLevel.equals("4") || packLevel.equals("8") || packLevel.equals("16"))
                    _packLevel = Integer.parseInt(packLevel);
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", packLevel, key));
            } else if (key.equals("com.ibm.etools.c.C_SIZE_OF_ENUM")) {
                String sizeEnum = (String) map.get(key);
                if (sizeEnum.equals("1") || sizeEnum.equals("2") || sizeEnum.equals("4") || sizeEnum.equals("5"))
                    _sizeEnums = Integer.parseInt(sizeEnum);
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", sizeEnum, key));
            } else if (key.equals("com.ibm.etools.c.C_SIZE_OF_WCHAR")) {
                String sizeWchar = (String) map.get(key);
                if (sizeWchar.equals("2") || sizeWchar.equals("4"))
                    _sizeWchar = Integer.parseInt(sizeWchar);
                else
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", sizeWchar, key));
            } else if (key.equals("com.ibm.etools.c.C_OPERATING_SYSTEM")) {
                _operatingSystem = (String) map.get(key);
            } else if (key.equals("com.ibm.etools.c.C_HARDWARE_PLATFORM")) {
                _hardwarePlatform = (String) map.get(key);
            } else if (key.equals("com.ibm.etools.c.C_COMPILER_VERSION")) {
                _compilerVersion = (String) map.get(key);
            } else if (key.equals("com.ibm.etools.c.C_OS_VERSION")) {
                _osVersion = (String) map.get(key);
            } else if (key.equals("com.ibm.etools.c.C_STRING_ENCODING")) {
                String stringEncoding = (String) map.get(key);
                if (stringEncoding != null && (stringEncoding.toUpperCase().equals(NULL) || stringEncoding.length() <= 0)) {
                    _isStringNullTerminated = true;

                } else {
                    _isStringNullTerminated = false;

                    if (stringEncoding != null && stringEncoding.toUpperCase().equals(SPACE)) {
                        _paddingCharacter = " ";

                    } else {
                        _paddingCharacter = stringEncoding;
                    }
                }
            } else if (key.equals("com.ibm.etools.c.C_PADDING_CHARACTER")) {
                _paddingCharacter = (String) map.get(key);
            } else if (key.equals("com.ibm.etools.c.C_FLOATING_POINT_FORMAT")) {
                _floatingPointFormat = (String) map.get(key);
            } else if (key.equals("com.ibm.etools.c.C_CODEPAGE")) {
                _codePage = (String) map.get(key);
                /*
                 * Validate the codepage before trying to use.
                 */
                try {
                    byte byteArray[] = { 12, 34 };
                    String s = new String(byteArray, _codePage);
                } catch (java.io.UnsupportedEncodingException ex) {
                    throw new CException(CResource.getString("_ERROR_Invalid_Value_for_Option", _codePage, key));
                }

            } else if (key.equals("com.ibm.etools.c.C_NO_TD")) {
                _isPopulateTD = false;
            } else if (key.equals("com.ibm.etools.c.C_NO_EXPAND")) {
                _isGenerateFullObjectList = false;
            } else if (key.equals("com.ibm.etools.c.C_REUSE_PARSE_INFO")) {
                _isReuseParseInfo = true;
            } else if (key.equals("com.ibm.etools.c.C_INCLUDE_GLOBAL_VARS")) {
                _isIncludeGlobalVariables = true;
            } else if (key.equals("com.ibm.etools.c.C_CREATE_FIELD_FOR_TYPE")) {
                _isCreateImplicitFieldsForTypes = true;
            } else if (key.equals("com.ibm.etools.c.C_NO_INCLUDES")) {
                _isProcessIncludeFiles = false;
            } else if (key.equals("com.ibm.etools.c.C_THROW_EXCEPTION_AT_PARSE_ERROR")) {
                _isThrowExceptionAtParseError = true;
            } else if (key.equals("com.ibm.etools.c.C_TREAT_UNKNOWN_TYPE_AS_ERROR")) {
                _isTreatUnknownTypeAsError = false;
            } else {
                throw new CException(CResource.getString("_ERROR_Invalid_Option", key));
            }

        }

        if (_alignment != PACKED_ALIGNMENT)
            _packLevel = 128;

    }

    /**
     * Reset all options to their default values.
     */
    public static void resetValues() {
        _compilerName = empty;
        _compilerFlags = empty;
        _includePaths = empty;

        _isBigEndian = false;
        _is64bitAddress = false;
        _isLongDouble128 = false;
        _isTreatUShortArrayAsDBCS = false;
        _alignment = NATURAL_ALIGNMENT;
        _packLevel = 128;
        _sizeEnums = 4;
        _sizeWchar = 2;

        _operatingSystem = null;
        _hardwarePlatform = null;
        _compilerVersion = null;
        _osVersion = null;
        _paddingCharacter = null;
        _floatingPointFormat = null;
        _codePage = null;
        _isStringNullTerminated = true;

        _isIncludeGlobalVariables = false;
        _isCreateImplicitFieldsForTypes = false;
        _isProcessIncludeFiles = true;
        _isGenerateFullObjectList = true;
        _isPopulateTD = true;
        _isReuseParseInfo = false;
        _isThrowExceptionAtParseError = false;
        _isTreatUnknownTypeAsError = false;
    }

    /**
     * Return the compiler name
     */
    public static String getCompilerName() {
        return _compilerName;
    }

    /**
     * Return the compiler flags
     */
    public static String getCompilerFlags() {
        return _compilerFlags;
    }

    /**
     * Return the include path
     */
    public static String getIncludePaths() {
        return _includePaths;
    }

    /**
     * Byte Order: Is it big endian?
     */
    public static boolean isBigEndian() {
        return _isBigEndian;
    }

    /**
     * Address Size: Is it 64bit?
     */
    public static boolean is64bitAddress() {
        return _is64bitAddress;
    }

    /**
     * Size of long double: Is it 128 bit?
     */
    public static boolean isLongDouble128() {
        return _isLongDouble128;
    }

    /**
     * Treat Unsigned Short Arrays as DBCS Strings?
     */
    public static boolean isTreatUShortArrayAsDBCS() {
        return _isTreatUShortArrayAsDBCS;
    }

    /**
     * Return the alignment rule
     */
    public static int getAlignment() {
        return _alignment;
    }

    /**
     * Return the pack level
     */
    public static int getPackLevel() {
        return _packLevel;
    }

    /**
     * Return the size of enums
     */
    public static int getSizeEnums() {
        return _sizeEnums;
    }

    /**
     * Return the size of wchar
     */
    public static int getSizeWchar() {
        return _sizeWchar;
    }

    /**
     * Return the operating system
     */
    public static String getOperatingSystem() {
        return _operatingSystem;
    }

    /**
     * Return the hardware platform
     */
    public static String getHardwarePlatform() {
        return _hardwarePlatform;
    }

    /**
     * Return the compiler version
     */
    public static String getCompilerVersion() {
        return _compilerVersion;
    }

    /**
     * Return the OS version
     */
    public static String getOSVersion() {
        return _osVersion;
    }

    /**
     * Is the string null terminated?
     */
    public static boolean isStringNullTerminated() {
        return _isStringNullTerminated;
    }

    /**
     * Return the padding character for a string
     */
    public static String getPaddingCharacter() {
        return _paddingCharacter;
    }

    /**
     * Return the floating point format
     */
    public static String getFloatingPointFormat() {
        return _floatingPointFormat;
    }

    /**
     * Return the code page
     */
    public static String getCodePage() {
        return _codePage;
    }

    /**
     * Are global variables included in the generated Extent?
     */
    public static boolean isIncludeGlobalVariables() {
        return _isIncludeGlobalVariables;
    }

    /**
     * Do we create implicit fields for types?
     */
    public static boolean isCreateImplicitFieldsForTypes() {
        return _isCreateImplicitFieldsForTypes;
    }

    /**
     * Do we process include files?
     */
    public static boolean isProcessIncludeFiles() {
        return _isProcessIncludeFiles;
    }

    /**
     * Do we generate full object list?
     */
    public static boolean isGenerateFullObjectList() {
        return _isGenerateFullObjectList;
    }

    /**
     * Do we populate the Type Descriptor model?
     */
    public static boolean isPopulateTypeDescriptorModel() {
        return _isPopulateTD;
    }

    /**
     * Do we reuse the parse information from the last importer session?
     */
    public static boolean isReuseParseInformation() {
        return _isReuseParseInfo;
    }

    /**
     * Throw an exception when we have a type parsing error?
     */
    public static boolean isThrowExceptionAtParseError() {
        return _isThrowExceptionAtParseError;
    }

    /**
     * Treat an unknown type as a parsing error?
     */
    public static boolean isTreatUnknownTypeAsError() {
        return _isTreatUnknownTypeAsError;
    }

}